# --Description: Use the GameSynth Tool API to render patches in AtomCraft project
import sys
import os
import cri.atomcraft.debug as acdebug
import cri.atomcraft.project as acproject
import socket

# --BeginUserVariable
"""
BITDEPTH:
  type:number
  comment:Bit-depth: 8, 16, 24, or 32-bit
CHANNELS:
  type:number
  comment:Number of channels: 1 or 2
DURATION:
  type:number
  comment:Duration used if infinite patch
VARIATIONS:
  type:number
  comment:Number of sounds to render
"""
BITDEPTH = 16
CHANNELS = 1
DURATION = 10
VARIATIONS = 1
# --EndUserVariable

#Connect to the GameSynth Tool API
try:
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	s.connect(("localhost", 28542))
	acdebug.log("*** Connected to the GameSynth Tool API ***")

except ConnectionRefusedError:
	acdebug.warning("Could not connect to GameSynth, please start GameSynth first.")
	sys.exit()

#Get the name of the current GameSynth patch
s.send("get_patchname".encode('utf-8'))
gs_patch_name = s.recv(28542).decode('utf-8')
gs_patch_name = gs_patch_name.rstrip("\r")
if "ERROR: no patch loaded" in gs_patch_name:
	acdebug.warning("GSAPI " + gs_patch_name)
	s.close()
	acdebug.log("*** Disconnected from the GameSynth Tool API ***")
	sys.exit()

# Find the current project and its absolute path
current_project = acproject.find_objects(None, "Project")["data"]
project_path = acproject.get_value(current_project[0], "ProjectFilePath")["data"]
project_path_root = os.path.dirname(project_path)

# Find the WorkUnits and its path from selection and check if the valid object is selected
object_types = ["WorkUnit", "CueSheet", "Cue", "Track"]

for object_type in object_types:
	selected_objects = acproject.get_selected_objects(object_type)["data"]

	if object_type == "WorkUnit" and selected_objects:	
		selected_workunits = acproject.get_selected_objects("WorkUnit")["data"]
		workunit_path = acproject.get_object_path(selected_workunits[0])["data"]
		break

	elif object_type in ["CueSheet", "Cue", "Track"] and selected_objects:
		selected_workunit = acproject.get_parent(selected_objects[0], "WorkUnit")["data"]
		selected_workunits = []
		selected_workunits.append(selected_workunit)
		workunit_path = acproject.get_object_path(selected_workunit)["data"]
		break

else:
	acdebug.warning(f"Select one of those objects: {object_types}.")
	s.close()
	acdebug.log("*** Disconnected from the GameSynth Tool API ***")
	sys.exit()

#Get the absolute Material path
materials_path = project_path_root + workunit_path + "/Materials"

#Create the GameSynth folder to render the patch if it doesn't already exist
gs_path = materials_path + "/GameSynth"
if not os.path.exists(gs_path):
	os.makedirs(gs_path)
	acdebug.log("GameSynth folder successfully created!")
else:
	acdebug.log("The GameSynth folder already exists.")

#---------- Function that builds the name of a wave file ---------
def get_wave_filename():
	global file_index
	#Get the path of the rendered file
	gs_wav_path = gs_path + "/" + gs_patch_name + "_" + str(file_index) + ".wav"
	base_name = gs_patch_name + "_" + str(file_index)

	# Increment the file counter for the next call
	file_index += 1
	return gs_wav_path, base_name
#-----------------------------------------------------------------

#----------------- Function that renders a patch -----------------
def render_patch():
	#Send the GameSynth Tool API command to render the patch
	gsapi_render_patch = f"render_patch \"{gs_wav_path}\" {BITDEPTH} {CHANNELS} {DURATION}"
	acdebug.log(gsapi_render_patch)

	render_patch = s.send(gsapi_render_patch.encode('utf-8'))
	render_patch_result = s.recv(28542).decode('utf-8')
	if "OK\r" not in render_patch_result:
		acdebug.warning("GSAPI " + render_patch_result.rstrip("\r"))
		s.close()
		acdebug.log("*** Disconnected from the GameSynth Tool API ***")
		sys.exit()
		return

	#Register a new Material
	material_root_folder = acproject.get_material_rootfolder(selected_workunits[0])["data"]
	acproject.register_unregistered_materials(material_root_folder) 

	#Build the lists of Materials and Material folders
	material_folder = acproject.get_child_object(material_root_folder, "MaterialSubFolder", "GameSynth")["data"]
	material = acproject.get_child_object(material_folder, "Material", base_name + ".wav")["data"]
	material_list.append(material)
	materials_folders_list.append(material_folder)
	return
#---------------------------------------------------------

#Check if the patch is infinite
s.send("is_infinite".encode('utf-8'))
is_infinite_result = s.recv(28542).decode('utf-8')
is_infinite_result = int(str(is_infinite_result).replace("\r", ""))
if ((is_infinite_result == 0) or (is_infinite_result == 1 and DURATION > 0)):
	file_index = 1
	base_name_list = []
	material_list = []
	materials_folders_list = []
	for _ in range(VARIATIONS):
		gs_wav_path, base_name = get_wave_filename()
		base_name_list.append(base_name)
		render_patch()
	#Sync with local files to refresh the Materials
	acproject.sync_local_files(materials_folders_list)
	
else:
	acdebug.warning("The patch is infinite, please set a duration higher than 0.")
	s.close()
	acdebug.log("*** Disconnected from the GameSynth Tool API ***")
	sys.exit()

#Close the TCP Connection
s.close()
acdebug.log("*** Disconnected from the GameSynth Tool API ***")

#------------------------------------------------------------------
#Create AtomCraft objects based on the selected object
selected_object_type = acproject.get_type_name(selected_objects[0]) ["data"]

#If a WorkUnit is selected
if selected_object_type not in ["CueSheet", "Cue", "Track"]:
	# ----- Creating Cue Sheet/Cue -----
	# Get the Cue Sheet Root folder
	cuesheet_rootfolder = acproject.get_cuesheet_rootfolder(selected_workunits[0])["data"]

	# Create a Cue Sheet Root folder named "GameSynth" if it does not exist yet
	cuesheet_folder = acproject.get_child_object(cuesheet_rootfolder, "CueSheetFolder", "GameSynth")["data"]
	if not cuesheet_folder:
		cuesheet_folder = acproject.create_object(cuesheet_rootfolder, "CueSheetFolder", "GameSynth")["data"]

	# Create a Cue Sheet in the Cue Sheet folder if it does not exist yet
	cuesheet = acproject.get_child_object(cuesheet_folder, "CueSheet", "GameSynth")["data"]
	if not cuesheet:
		cuesheet = acproject.create_object(cuesheet_folder, "CueSheet", "GameSynth")["data"]

#If a CueSheet is selected
if selected_object_type not in ["Cue", "Track"]:
	# ----- Creating Cue -----
	if selected_object_type == "CueSheet":
		cuesheet = selected_objects[0]

	# Create a Cue unless it already exists
	cue = acproject.get_child_object(cuesheet, "Cue", gs_patch_name)["data"]
	if not cue:
		cue = acproject.create_object(cuesheet, "Cue", gs_patch_name)["data"]

	# Change the type of the Cue if there are several variations
	if VARIATIONS > 1:
		acproject.set_value(cue, "SequenceType", "RandomNoRepeat")

#If a Cue is selected
if selected_object_type not in ["Track"]:
	# ----- Creating Track(s) -----
	if selected_object_type == "Cue":
		cue = selected_objects[0]

	# Get the Track name
	for track_index in range(1, VARIATIONS + 1):
		track_name = "Track_" + gs_patch_name + "_" + str(track_index)

		# Create a Track unless it already exists
		track = acproject.get_child_object(cue, "Track", track_name)["data"]
		if not track:
			track = acproject.create_object(cue, "Track", track_name)["data"]
		
		# ----- Creating Waveform Region(s) -----
		# Create a waveform region by specifying the Material, unless it already exists
		material_index = (track_index - 1) % VARIATIONS # Adjust the material index to start from 0
		material_name = acproject.get_value(material_list[material_index], "Name")["data"]
		waveform_region = acproject.get_child_object(track, "WaveformRegion", material_name)["data"]
		if not waveform_region:
			waveform_region = acproject.create_waveform_region(track, material_list[material_index])["data"]

if selected_object_type == "Track":
	track = selected_objects[0]
	# ----- Creating Waveform Region if VARIATIONS is 1 -----
	if VARIATIONS == 1:
		material_name = acproject.get_value(material_list[0], "Name")["data"]
		waveform_region = acproject.get_child_object(track, "WaveformRegion", material_name)["data"]
		if not waveform_region:
			waveform_region = acproject.create_waveform_region(track, material_list[0])["data"]
	# ----- Creating SubSequence -----
	else:
		track_name = acproject.get_value(track, "Name")["data"]
		sub_sequence = acproject.get_child_object(track, "SubSequence", gs_patch_name)["data"]
		if not sub_sequence:
			sub_sequence = acproject.create_object(track, "SubSequence", gs_patch_name)["data"]
		# Change the type of SubSequence if there are several variations
		if VARIATIONS > 1:
			acproject.set_value(sub_sequence, "SequenceType", "RandomNoRepeat")

		# Get the Track name
		for track_index in range(1, VARIATIONS + 1):
			track_name = "Track_" + gs_patch_name + "_" + str(track_index)

			# Create a Track unless it already exists
			track = acproject.get_child_object(sub_sequence, "Track", track_name)["data"]
			if not track:
				track = acproject.create_object(sub_sequence, "Track", track_name)["data"]
		
			# ----- Creating Waveform Region(s) -----
			# Create a waveform region by specifying a Material unless it already exists
			material_index = (track_index - 1) % VARIATIONS # Adjust the material index to start from 0
			material_name = acproject.get_value(material_list[material_index], "Name")["data"]
			waveform_region = acproject.get_child_object(track, "WaveformRegion", material_name)["data"]
			if not waveform_region:
				waveform_region = acproject.create_waveform_region(track, material_list[material_index])["data"]
